/***************************************************************************
                          midifile.cpp  -  description
                             -------------------
    begin                : Sun Nov 10 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : morten@debian.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#undef Q_DISABLE_COPY
#include <qfile.h>
#include <qstring.h>
#include <kdebug.h>

#include "midifile.h"

MIDIFile::MIDIFile() : QFile()
{
}

MIDIFile::MIDIFile(const QString& name) : QFile(name)
{
}

MIDIFile::~MIDIFile()
{
}

const QPtrVector<MIDIFile::Chunk>&
MIDIFile::chunks()
{
  if(_chunks.isEmpty())
  {
    Offset oldOffset = at();
    Offset remaining = size();

    at(0);

    while(remaining)
    {
      char id[4];

      readBlock(id, 4);

      uint size;
      size = getch() << 24;
      size |= getch() << 16;
      size |= getch() << 8;
      size |= getch();

      remaining -= 8;

      if(size > remaining)
      {
        kdWarning() << "Invalid chunk size in SMF file `" << name()
                    << "' at offset " << (at() - 8) << " in chunk with ID `"
                    << id[0] << id[1] << id[2] << id[3] << "'." << endl;
        kdWarning() << "Chunk size is " << size << " but there are only "
                    << remaining << " bytes left in the file." << endl;

        break;
      }

      Chunk* chunk = new Chunk(id, size);

      remaining -= size;

      while(size--)
        chunk->put(getch());

      _chunks.resize(_chunks.size() + 1);
      _chunks.insert(_chunks.size() - 1, chunk);
    }

    at(oldOffset);
  }

  return _chunks;
}

int
MIDIFile::writeChunk(const MIDIFile::Chunk& chunk)
{
  if(4 != writeBlock(chunk.id().latin1(), 4))
    return -1;

  uint size = chunk.size();

#define writeChunk_TEST(x) do { if(-1 == (x)) return -1; } while(0)

  writeChunk_TEST(putch(static_cast<unsigned char>(size >> 24)));
  writeChunk_TEST(putch(static_cast<unsigned char>(size >> 16)));
  writeChunk_TEST(putch(static_cast<unsigned char>(size >> 8)));
  writeChunk_TEST(putch(static_cast<unsigned char>(size)));

  for(uint i = 0; i < size; ++i)
    writeChunk_TEST(putch(chunk.data(i)));

  return 8 /* header size*/ + size /* data size */;
}

MIDIFile::Chunk::Chunk()
{
}

MIDIFile::Chunk::Chunk(const char* id, uint size)
{
  _id[0] = id[0];
  _id[1] = id[1];
  _id[2] = id[2];
  _id[3] = id[3];
  _data.reserve(size);
}

MIDIFile::Chunk::~Chunk()
{
}

QString
MIDIFile::Chunk::id() const
{
  QString ret;

  ret.setLength(4);

  for(uint i = 0; i < 4; ++i)
    ret[i] = _id[i];

  return ret;
}

uint 
MIDIFile::Chunk::size() const
{
  return _data.size();
}

unsigned char
MIDIFile::Chunk::data(uint offset) const
{
  return _data[offset];
}

void
MIDIFile::Chunk::put(unsigned char data)
{
  _data.push_back(data);
}

MIDIFile::MThd::MThd(Format format, uint tracks, uint ppqn)
  : Chunk("MThd", 6)
{
  _data.push_back(format >> 8);
  _data.push_back(format);
  _data.push_back(tracks >> 8);
  _data.push_back(tracks);
  _data.push_back(ppqn >> 8);
  _data.push_back(ppqn);
}

MIDIFile::MThd::Format
MIDIFile::MThd::format() const
{
  return static_cast<Format>(_data[0] << 8 | _data[1]);
}

uint
MIDIFile::MThd::tracks() const
{
  return (_data[2] << 8) | _data[3];
}

uint
MIDIFile::MThd::ppqn() const
{
  return (_data[4] << 8) | _data[5];
}

MIDIFile::MTrk::MTrk()
  : Chunk("MTrk", 3),
    _time(0)
{
}

MIDIFile::MTrk::ConstIterator
MIDIFile::MTrk::begin() const
{
  return ConstIterator(_data.begin());
}

MIDIFile::MTrk::ConstIterator
MIDIFile::MTrk::end() const
{
  return ConstIterator(_data.end());
}

MIDIFile::MTrk::Event::Event()
{
}

MIDIFile::MTrk::Event::Event(uint time)
{
  _time = time;
}

uint
MIDIFile::MTrk::Event::time() const
{
  return _time;
}

uint 
MIDIFile::MTrk::Event::size() const
{
  return _data.size();
}

unsigned char
MIDIFile::MTrk::Event::data(uint offset) const
{
  return _data[offset];
}

unsigned char
MIDIFile::MTrk::Event::put(unsigned char data)
{
  _data.push_back(data);

  return data;
}

void
MIDIFile::MTrk::Event::clear()
{
  _data.clear();
}

MIDIFile::MTrk::ConstIterator::ConstIterator()
{
}

MIDIFile::MTrk::ConstIterator::ConstIterator(const ConstIterator& it)
  : _pos(it._pos),
    _time(it._time),
    _length(it._length),
    _status(it._status),
    _fetched(it._fetched)
{
}

MIDIFile::MTrk::ConstIterator::ConstIterator(QValueVector<unsigned char>::const_iterator begin)
  : _pos(begin),
    _time(0),
    _fetched(false)
{
}

MIDIFile::MTrk::Event
MIDIFile::MTrk::ConstIterator::operator*()
{
  if(!_fetched)
    fetch();

  Event ret(_time);
  ret.put(_status);

  QValueVector<unsigned char>::const_iterator pos = _pos;

  for(uint i = _length; i--; )
    ret.put(*pos++);
  
  return ret;
}

MIDIFile::MTrk::ConstIterator&
MIDIFile::MTrk::ConstIterator::operator++()
{
  if(!_fetched)
    fetch();

  _pos += _length;
  _fetched = false;

  return *this;
}

MIDIFile::MTrk::ConstIterator
MIDIFile::MTrk::ConstIterator::operator++(int)
{
  if(!_fetched)
    fetch();

  ConstIterator ret = *this;

  _pos += _length;
  _fetched = false;

  return ret;
}

bool
MIDIFile::MTrk::ConstIterator::operator==(const MIDIFile::MTrk::ConstIterator::ConstIterator& it) const
{
  return _pos == it._pos;
}

bool
MIDIFile::MTrk::ConstIterator::operator!=(const MIDIFile::MTrk::ConstIterator::ConstIterator& it) const
{
  return _pos != it._pos;
}

void MIDIFile::MTrk::ConstIterator::fetch()
{
  uint relTime = 0;

  do
  {
    relTime = (relTime << 7) | (*_pos & 0x7F);
  }
  while(*_pos++ & 0x80);

  _time += relTime;

  if(*_pos & 0x80)
  {
    _status = *_pos++;
  }

  switch(_status & 0xF0)
  {
  case 0x80:
  case 0x90:
  case 0xA0:
  case 0xB0:
  case 0xE0:

    _length = 2;

    break;

  case 0xC0:
  case 0xD0:

    _length = 1;

    break;

  case 0xF0:

    if(_status == 0xFF) // Meta-event
      _status = *_pos++;

    _length = 0;

    do
    {
      _length = (_length << 7) | (*_pos & 0x7F);
    }
    while(*_pos++ & 0x80);
  }

  _fetched = true;
}

void MIDIFile::MTrk::put(const MIDIFile::MTrk::Event& event)
{
  uint deltaTime = event.time() - _time;
 
  bool p = false;
  if((deltaTime >> 21) != 0) Chunk::put((deltaTime >> 21) | 0x80), ++p;
  if((deltaTime >> 14) != 0 || p) Chunk::put((deltaTime >> 14) | 0x80), ++p;
  if((deltaTime >> 7) != 0 || p) Chunk::put((deltaTime >> 7) | 0x80);
  Chunk::put(deltaTime & 0x7F);

  for(uint i = 0; i < event.size(); ++i)
    Chunk::put(event.data(i));

  _time = event.time();
}

// vim: ts=2 sw=2 et
