/***************************************************************************
                          midifile.h  -  handler for SMF files
                             -------------------
    begin                : Sun Nov 10 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : morten@debian.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef MIDIFILE_H
#define MIDIFILE_H

#include <qfile.h>
#include <qglobal.h>
#include <qptrvector.h>
#include <qvaluevector.h>

class QString;

/**
 * The MIDIFile class provides methods for accessing files in the Standard
 * MIDI Format (SMF).
 *
 * The Standard MIDI Format is a chunk based extensible format, supporting
 * extensions in the form of vendor-specific chunks. This format is used
 * by almost all MIDI software, and is generally accepted as the only
 * format you will ever need for storing MIDI data.
 *
 * This class provides read only and write only access to SMF files, and
 * provides convenience subclasses for MThd and MTrk chunks (header and
 * data, respectively). It is a subclass of QFile, and you can expect
 * the file position pointers to be kept valid after a call to any function
 * in this class.
 */
class MIDIFile : public QFile
{
public:

  /**
   * Constructs a MIDI file with no name.
   */
  MIDIFile();

  /**
   * Constructs a MIDI file with the name @p name.
   *
   * @see setName
   */
  MIDIFile(const QString& name);

  /**
   * Destroys a MIDI file. Calls @ref QFile::close().
   */
  ~MIDIFile();

  /**
   * The MIDIFile::Chunk class represents a chunk in a SMF file.
   *
   * When creating a new MIDI file, create the chunks one by one, fill
   * them with data, and write them using @ref MIDIFile::writeChunk().
   */
  class Chunk
  {
  public:

    /**
     * Constructs an empty chunk.
     */

    Chunk();

    /**
     * Constructs a chunk with id @p id and size @p size.
     */
    Chunk(const char* id, uint size);

    /**
     * Destroys the chunk object.
     */
    ~Chunk();

    /**
     * Returns the ID of the chunk.
     */
    QString id() const;

    /**
     * Returns the size of the data part of the chunk in bytes.
     *
     * @see data
     */
    uint size() const;

    /**
     * Returns the byte at offset @p offset in the chunk.
     */
    unsigned char data(uint offset) const;

    /**
     * Adds a byte of data to the end of the cunk.
     */
    void put(unsigned char data);

  protected:

    char _id[4];

    QValueVector<unsigned char> _data;
  };

  /**
   * ChunkVector is a convenience typedef.
   */
  typedef QPtrVector<Chunk> ChunkVector;

  /**
   * The MThd class represents the header chunk present in all SMF files.
   *
   * If the id of a @ref Chunk is "MThd", you can safely cast it to an MThd.
   */
  class MThd : public Chunk
  {
  public:

    /**
     * Formats:
     *
     * @li SingleTrack - All MIDI data of all 16 channles is contained in
     *                   on track.
     *
     * @li SimultaneousTracks - All tracks start at time 0
     *
     * @li Patterns - Tracks are sequentially independent
     */
    enum Format
    {
      SingleTrack = 0,
      SimultaneousTracks = 1,
      Patterns = 2
    };

    /**
     * Creates a Standard MIDI Format header chunk, with format @p format,
     * @p numTracks tracks and @p ppqn pulses per quarter note.
     */
    MThd(Format format = SingleTrack, uint numTracks = 1, uint ppqn = 24);

    /**
     * Returns the format of the file this chunk belongs to.
     */
    Format format() const;

    /**
     * Returns the number of tracks in the file this chunks belongs to.
     */
    uint tracks() const;

    /**
     * Returns the amount of pulses (MIDI clocks) per quarter note.
     */
    uint ppqn() const;
  };

  /**
   * The MTrk class represents the data chunk of SMF files.
   *
   * If the id of a @ref Chunk is "MTrk", you can safely cast it to an MThd.
   */
  class MTrk : public Chunk
  {
  public:

    /**
     * Constructs an empty MTrk chunk.
     */
    MTrk();

    /**
     * The Event class represents an event in an MTrk chunk.
     *
     * Each event contains exactly one MIDI event with a time stamp.  This
     * class handles running status, so the caller need not think of that.
     * Events with first byte less than 0x80 are meta-events.
     */
    class Event
    {
    public:

      /**
       * Constructs an empty event.
       */
      Event();

      /**
       * Constructs an event for the time @p time.
       *
       * @see put
       */
      Event(uint time);

      /**
       * Returns the time of this event.
       */
      uint time() const;

      /**
       * Returns the size of the data part of this event in bytes.
       *
       * @see data
       */
      uint size() const;

      /**
       * Returns the byte at offset @p offset in this event.
       */
      unsigned char data(uint offset) const;

      /**
       * Adds a byte of data at the end of this event. Returns @p data.
       */
      unsigned char put(unsigned char data);

      /**
       * Clears the data of this event. Useful for reusing an Event object
       * for multiple events with same timestamp.
       */
      void clear();

    protected:

      QValueVector<unsigned char> _data;
      uint _time;
    };

    class ConstIterator
    {
    public:

      /**
       * Constructs an uninitialized iterator.
       */
      ConstIterator();

      /**
       * Constructs an copy of the iterator @p it.
       */
      ConstIterator(const ConstIterator& it);

      /**
       * Asterisk operator. Returns a copy of the current event.
       */
      Event operator*();

      /**
       * Prefix ++ makes the succeeding event current and returns an iterator
       * pointing to the new current event. The iterator cannot check whether
       * it reached the end of the list. Incrementing the iterator returned
       * by end() causes undefined results.
       */
      ConstIterator& operator++();

      /**
       * Postfix ++ makes the succeeding event current and returns an iterator
       * pointing to the old current event. The iterator cannot check whether
       * it reached the end of the list. Incrementing the iterator returned
       * by end() causes undefined results.
       */
      ConstIterator operator++(int);

      /**
       * Compares this iterator to @p it and returns TRUE if they point to the
       * same item; otherwise returns FALSE.
       */
      bool operator==(const ConstIterator& it) const;

      /**
       * Compares this iterator to @p it and returns TRUE if they point to
       * different items; otherwise returns FALSE.
       */
      bool operator!=(const ConstIterator& it) const;

    private:

      friend class MTrk;

      ConstIterator(QValueVector<unsigned char>::const_iterator begin);

    protected:

      QValueVector<unsigned char>::const_iterator _pos;
      uint _time;
      uint _length;
      unsigned char _status;
      bool _fetched;

      void fetch();
    };

#define METAEVENT_BEGIN(MetaEvent, Parms) \
    class MetaEvent : public Event \
    { \
    public: \
      MetaEvent Parms : Event(time) \
      { \
        put(0xFF);

#define METAEVENT_STRING(string) \
    do { \
      size_t length = strlen(string); \
      bool p = FALSE; \
      if((length >> 21) != 0) put((length >> 21) | 0x80), ++p; \
      if((length >> 14) != 0 || p) put((length >> 14) | 0x80), ++p; \
      if((length >> 7) != 0 || p) put((length >> 7) | 0x80); \
      put(length & 0x7F); \
      for(uint i = 0; i < length; ++i) \
        put(string[i]); \
    } while (0)

#define METAEVENT_END \
      } \
    };

    METAEVENT_BEGIN(SequenceNumber, (uint cue, uint time = 0))
      if(cue)
      {
        put(0x00);
        put(0x02);
      }
      put(cue >> 8);
      put(cue);
    METAEVENT_END

    METAEVENT_BEGIN(Text, (const char* text, uint time))
      put(0x01);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(Copyright, (const char* text, uint time))
      put(0x02);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(TrackName, (const char* text, uint time = 0))
      put(0x03);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(Instrument, (const char* text, uint time = 0))
      put(0x04);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(Lyric, (const char* syllable, uint time))
      put(0x05);
      METAEVENT_STRING(syllable);
    METAEVENT_END

    METAEVENT_BEGIN(Marker, (const char* text, uint time))
      put(0x06);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(CuePoint, (const char* text, uint time))
      put(0x07);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(PatchName, (const char* text, uint time))
      put(0x08);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(PortName, (const char* text, uint time = 0))
      put(0x09);
      METAEVENT_STRING(text);
    METAEVENT_END

    METAEVENT_BEGIN(EndOfTrack, (uint time))
      put(0x2F);
      put(0x00);
    METAEVENT_END

    METAEVENT_BEGIN(Tempo, (uint mspq, uint time = 0))
      put(0x51);
      put(0x03);
      put(mspq >> 16);
      put(mspq >> 8);
      put(mspq);
    METAEVENT_END

    METAEVENT_BEGIN(SMPTEOffset, (uint hour, uint minute, uint second, uint frame, uint centiframe, uint time = 0))
      put(0x54);
      put(0x05);
      put(hour);
      put(minute);
      put(second);
      put(frame);
      put(centiframe);
    METAEVENT_END

    METAEVENT_BEGIN(TimeSignature, (uint numerator, uint denominator, uint clocks, uint freq, uint time = 0))
      put(0x58);
      put(0x04);
      put(numerator);
      put(denominator);
      put(clocks);
      put(freq);
    METAEVENT_END

    METAEVENT_BEGIN(KeySignature, (signed char sharps, bool major, uint time = 0))
      put(0x59);
      put(0x02);
      put(static_cast<unsigned char>(sharps));
      put(major ? 0 : 1);
    METAEVENT_END

    METAEVENT_BEGIN(Proprietary, (const unsigned char* data, uint length, uint time = 0))
      put(0x7F);
      bool p = FALSE;
      if((length >> 21) != 0) put((length >> 21) | 0x80), ++p;
      if((length >> 14) != 0 || p) put((length >> 14) | 0x80), ++p;
      if((length >> 7) != 0 || p) put((length >> 7) | 0x80);
      put(length & 0x7F);
      for(uint i = 0; i < length; ++i)
        put(data[i]);
    METAEVENT_END

    typedef ConstIterator const_iterator;

    ConstIterator begin() const;
    ConstIterator end() const;

    /**
     * Adds the event @p event to this chunk.
     */
    void put(const Event& event);

  protected:

    uint _time;
  };

  /**
   * Returns all the chunks in the MIDI file.  The chunks are only
   * read from the file the first time this function is called.
   */
  const ChunkVector& chunks();

  /**
   * Writes the chunk @p chunk to the file.  Note that all SMF files
   * must begin with a MThd chunk.
   *
   * Returns the amount of bytes written, or -1 if an error occrs.
   */
  int writeChunk(const Chunk& chunk);

protected:

  ChunkVector _chunks;
};

#endif // !MIDIFILE_H

// vim: ts=2 sw=2 et
