/***************************************************************************
                          pattern.cpp  -  Pattern view
                             -------------------
    begin                : Sat Jun 8 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qlabel.h>
#include <qcheckbox.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qpainter.h>
#include <qscrollbar.h>
#include <qspinbox.h>

// include files for KDE
#include <kdebug.h>
#include <klocale.h>
#include <kpopupmenu.h>

#include "clipboard.h"
#include "configuration.h"
#include "midi.h"
#include "mididriver.h"
#include "pattern.h"
#include "sanetrackerdoc.h"
#include "scheduler.h"

const int skip = 3;

Pattern::Pattern(QWidget* parent, SanetrackerDoc& doc, const Configuration* config, Scheduler& scheduler)
  : QWidget(parent, "Pattern View"),
    doc(doc),
    scheduler(scheduler),
    device(~0),
    pattern(0),
    position(0),
    row(0),
    col(0)
{
  this->config = config;

  if(!fontPixmaps[0][0])
  {
    char background[12];
    char backgroundDark[12];
    char highlight[12];

    sprintf(background, " \tc %s", colorGroup().background().name().ascii());
    sprintf(backgroundDark, " \tc %s", colorGroup().background().dark(110).name().ascii());
    sprintf(highlight, " \tc %s", colorGroup().highlight().name().ascii());

    for(int i = 0; i < 25; ++i)
    {
      const char* charBackground[10] =
      {
        font[i][0], background, font[i][1], font[i][2], font[i][3],
        font[i][4], font[i][5], font[i][6], font[i][7], font[i][8],
      };

      const char* charBackgroundDark[10] =
      {
        font[i][0], backgroundDark, font[i][1], font[i][2], font[i][3],
        font[i][4], font[i][5], font[i][6], font[i][7], font[i][8],
      };

      const char* charHighlight[10] =
      {
        font[i][0], highlight, font[i][1], font[i][2], font[i][3],
        font[i][4], font[i][5], font[i][6], font[i][7], font[i][8],
      };

      fontPixmaps[Background][i] = new QPixmap(charBackground);
      fontPixmaps[BackgroundDark][i] = new QPixmap(charBackgroundDark);
      fontPixmaps[Highlight][i] = new QPixmap(charHighlight);
    }
  }

  vLayout = new QVBoxLayout(this);
  hLayout = new QHBoxLayout(vLayout);

  labelPatch = new QLabel(i18n("&Patch:"), this);
  comboPatches = new QComboBox(this);
  comboPatches->setFocusPolicy(QWidget::ClickFocus);
  labelStep = new QLabel(i18n("S&tep:"), this);
  spinStep = new QSpinBox(this);
  spinStep->setFocusPolicy(QWidget::ClickFocus);
  spinStep->setMinValue(1);
  spinStep->setMaxValue(16);
  checkContinuous = new QCheckBox(i18n("&Continuous editing"), this);
  checkContinuous->setFocusPolicy(QWidget::ClickFocus);
  data = new Data(this);
  scrollHoriz = new QScrollBar(this);
  scrollHoriz->setOrientation(Qt::Horizontal);
  scrollHoriz->setMaxValue(0);

  labelPatch->setBuddy(comboPatches);
  labelStep->setBuddy(spinStep);

  comboPatches->setMinimumWidth(150);

  menu = new KPopupMenu(data, "Sequence menu");
  menu->insertItem(i18n("&Clear track"), this, SLOT(clearTrack()));
  menu->insertItem(i18n("C&opy track"), this, SLOT(copyTrack()));
  menu->insertItem(i18n("&Paste track"), this, SLOT(pasteTrack()));
  menu->insertItem(i18n("&Transpose track +12"), this,
                   SLOT(transposeTrackOctaveUp()));
  menu->insertItem(i18n("T&ranspose track -12"), this,
                   SLOT(transposeTrackOctaveDown()));
  menu->insertSeparator();
  menu->insertItem(i18n("C&lear pattern"), this, SLOT(clearPattern()));
  menu->insertItem(i18n("Cop&y pattern"), this, SLOT(copyPattern()));
  menu->insertItem(i18n("P&aste pattern"), this, SLOT(pastePattern()));
  menu->insertItem(i18n("Tra&nspose pattern +12"), this,
                   SLOT(transposePatternOctaveUp()));
  menu->insertItem(i18n("Transpo&se pattern -12"), this,
                   SLOT(transposePatternOctaveDown()));

  hLayout->setMargin(2);
  hLayout->addWidget(labelPatch);
  hLayout->addWidget(comboPatches);
  hLayout->addWidget(labelStep);
  hLayout->addWidget(spinStep);
  hLayout->addWidget(checkContinuous);
  hLayout->addStretch();
  vLayout->addWidget(data);
  vLayout->addWidget(scrollHoriz);

  connect(comboPatches, SIGNAL(activated(int)), this, SLOT(patchActivated(int)));

  edit(0, 0, 0);
}

Pattern::~Pattern()
{
}

void Pattern::setActive(bool active)
{
  if(active == _active)
    return;

  _active = active;

  if(doc.size())
    data->updateCell(row, col);
}

void Pattern::keyPressed(uint note, uint velocity)
{
  if(activeNotes.find(note) != activeNotes.end())
    return;

  if(active() && pattern)
  {
    SanetrackerDoc::Pattern::Track& track = doc[device][pattern][col + activeNotes.size()];
  
    track[row * skip].setNote(note);
    track[row * skip].setRelativity(0);
    track[row * skip].setVelocity(velocity);
      
    data->updateCell(row, col + activeNotes.size());

    doc.setModified();
  }

  MIDIDriver::putRT(doc[device].portIndex(),
                    MIDI_NOTE_ON(doc[device].deviceID()));
  MIDIDriver::putRT(doc[device].portIndex(), note);
  MIDIDriver::putRT(doc[device].portIndex(), velocity);
  
  activeNotes.append(note);
}

void Pattern::keyReleased(uint note, uint velocity)
{
  QValueList<uint>::iterator i;
  
  if((i = activeNotes.find(note)) == activeNotes.end())
    return;

  MIDIDriver::putRT(doc[device].portIndex(), MIDI_NOTE_OFF(doc[device].deviceID()));
  MIDIDriver::putRT(doc[device].portIndex(), note);
  MIDIDriver::putRT(doc[device].portIndex(), velocity);

  activeNotes.remove(i);
  
  if(active() && pattern && activeNotes.empty())
  {
    data->scroll(spinStep->value());
  }
}

void Pattern::clearTrack()
{
  doc[device][pattern][selectCol] = SanetrackerDoc::Pattern::Track();
}

void Pattern::copyTrack()
{
  Clipboard::track = doc[device][pattern][selectCol];
}

void Pattern::pasteTrack()
{
  doc[device][pattern][selectCol] = Clipboard::track;
}

void Pattern::transposeTrackOctaveUp()
{
  uint measureLength = doc.measureLength(position);

  for(uint i = 0; i < measureLength; ++i)
  {
    SanetrackerDoc::Pattern::Track::Cell& cell =
      doc[device][pattern][selectCol][i];

    if(cell.isNote())
    {
      if(cell.note() > 127 - 12)
        cell.clear();
      else
        doc[device][pattern][selectCol][i].setNote(
          doc[device][pattern][selectCol][i].note() + 12);
    }
  }
}

void Pattern::transposeTrackOctaveDown()
{
  uint measureLength = doc.measureLength(position);

  for(uint i = 0; i < measureLength; ++i)
  {
    SanetrackerDoc::Pattern::Track::Cell& cell =
      doc[device][pattern][selectCol][i];

    if(cell.isNote())
    {
      if(cell.note() < 12)
        cell.clear();
      else
        doc[device][pattern][selectCol][i].setNote(
          doc[device][pattern][selectCol][i].note() - 12);
    }
  }
}

void Pattern::clearPattern()
{
  doc[device][pattern] = SanetrackerDoc::Pattern();
}

void Pattern::copyPattern()
{
  Clipboard::pattern = doc[device][pattern];
}

void Pattern::pastePattern()
{
  doc[device][pattern] = Clipboard::pattern;
}

void Pattern::transposePatternOctaveUp()
{
  uint measureLength = doc.measureLength(position);

  for(uint j = 0; j < doc[device][pattern].size(); ++j)
  {
    for(uint i = 0; i < measureLength; ++i)
    {
      SanetrackerDoc::Pattern::Track::Cell& cell =
        doc[device][pattern][j][i];

      if(cell.isNote())
      {
        if(cell.note() < 12)
          cell.clear();
        else
          doc[device][pattern][j][i].setNote(
            doc[device][pattern][j][i].note() - 12);
      }
    }
  }
}

void Pattern::transposePatternOctaveDown()
{
  uint measureLength = doc.measureLength(position);

  for(uint j = 0; j < doc[device][pattern].size(); ++j)
  {
    for(uint i = 0; i < measureLength; ++i)
    {
      SanetrackerDoc::Pattern::Track::Cell& cell =
        doc[device][pattern][j][i];

      if(cell.isNote())
      {
        if(cell.note() > 127 - 12)
          cell.clear();
        else
          doc[device][pattern][j][i].setNote(
            doc[device][pattern][j][i].note() + 12);
      }
    }
  }
}

void Pattern::patchActivated(int patch)
{
  if(pattern && doc[device][pattern].patch() != static_cast<uint>(patch))
  {
    doc[device][pattern].setPatch(patch);

    doc.setModified();
  }

  if(!scheduler.isPlaying())
    scheduler.setPatch(device, patch);
}

void Pattern::edit(uint _device, uint _pattern, uint _position)
{
  if(_device == device && _pattern == pattern)
    return;
 
  if(_device >= doc.size())
  {
    pattern = 0;

    return;
  }

  if(_device != device)
  {
    const Configuration::Instrument& instr = config->instrument(doc[_device].configType());

    comboPatches->clear();

    comboPatches->insertItem("(Current)");

    int n = 1;
    QMap<int, QString>::ConstIterator i;

    for(i = instr.banks().begin(); i != instr.banks().end(); ++i, ++n)
    {
      if(i.key() == -1)
        continue;

      const Configuration::Bank& bank = config->bank(i.data());

      Configuration::Bank::ConstIterator j;

      for(j = bank.begin(); j != bank.end(); ++j)
        comboPatches->insertItem(j.data());
    }

    device = _device;
  }

  pattern = _pattern;

  if(pattern == 0)
  {
    data->update();
    
    return;
  }

  position = _position;

  comboPatches->setCurrentItem(doc[device][pattern].patch());
  patchActivated(doc[device][pattern].patch());

  subFocus = 0;

  data->update();
}

void Pattern::drawText(QPainter& painter, int x, int y, const QString& string, TextMode mode)
{
  const char* chars = "0123456789ABCDEFG-+=#b[_]";

  for(uint n = 0; n < string.length(); ++n)
  {
    char* pos = strchr(chars, string[n]);

    if(pos)
      painter.drawPixmap(n * 8 + x, y, *fontPixmaps[static_cast<int>(mode)][pos - chars]);
  }
}

Pattern::Data::Data(Pattern* parent) : QWidget(parent)
{
  this->parent = parent;

  setFocusPolicy(QWidget::StrongFocus);

  setWFlags(Qt::WRepaintNoErase);
}

void Pattern::Data::updateCell(uint row, uint col)
{
  int x = 2 * 7 + 11 + col * (8 * 7 + 5);

  int middle = height() / 2;
  int i = row - parent->row;

  if(i < 0)
  {
    update(x, middle - 11 + i * 8, 8 * 7, 8);
  }
  else if(i == 0)
  {
    update(x, middle - 6, 8 * 7, 11);
  }
  else // i > 0
  {
    update(x, middle + 3 + i * 8, 8 * 7, 8);
  }
}

void Pattern::Data::scroll(int rows)
{
  if(rows == 0)
    return;

  uint patternLength = parent->doc.measureLength(parent->position) / skip;

  if(!(rows < 0 && static_cast<uint>(-rows) > parent->row)
  && static_cast<uint>(parent->row + rows) < patternLength)
  {
    parent->row += rows;

    int middle = height() / 2;

    QWidget::scroll(0, -rows * 8, QRect(4, 3, width() - 7, middle - 11 - 3));
    QWidget::scroll(0, -rows * 8, QRect(4, middle + 11, width() - 7, middle - 14));
    update(4, middle - 6, width() - 7, 11);
  }
  else // !(0 <= row + rows < patternLength)
  {
    if(rows < 0)
    {
      if(parent->checkContinuous->isChecked())
        emit parent->prevPattern();

      parent->row += (patternLength + rows);
    }
    else // rows > 0
    {
      if(parent->checkContinuous->isChecked())
        emit parent->nextPattern();

      parent->row -= (patternLength - rows);
    }

    update();
  }
}

void Pattern::keyPressed(QKeyEvent* event)
{
  if(!pattern)
    return;

  if(subFocus == 2)
  {
    const char* hex = "01234567";

    for(int i = 0; i < 8; ++i)
    {
      if(event->key() == hex[i])
      {
        if(doc[device][pattern][col][row * skip].isNote())
           doc[device][pattern][col][row * skip].setVelocity(
           (doc[device][pattern][col][row * skip].velocity() & 0x0F) | (i << 4));

        doc.setModified();

        data->scroll(spinStep->value());

        event->accept();

        return;
      }
    }
  }
  else if(subFocus == 3)
  {
    const char* hex = "0123456789ABCDEF";

    for(int i = 0; i < 16; ++i)
    {
      if(event->key() == hex[i])
      {
        if(doc[device][pattern][col][row * skip].isNote())
           doc[device][pattern][col][row * skip].setVelocity(
           (doc[device][pattern][col][row * skip].velocity() & 0x70) | i);

        doc.setModified();

        data->scroll(spinStep->value());

        event->accept();

        return;
      }
    }
  }

  switch(event->key())
  {
  case Key_Escape:

    // XXX: lose focus

    break;

  case Key_A:

    doc[device][pattern][col][row * skip].setStop();

    doc.setModified();

    data->scroll(spinStep->value());

    break;

  case Key_Up:

    data->scroll(-1);

    break;

  case Key_Down:

    data->scroll(1);

    break;

  case Key_Left:

    if(subFocus == 0)
    {
      if(col == 0)
        break;

      subFocus = 4;

      --col;

      data->updateCell(row, col + 1);
    }

    subFocus--;

    data->updateCell(row, col);

    break;

  case Key_Right:

    subFocus++;

    if(subFocus == 4)
    {
      subFocus = 0;

      ++col;

      data->updateCell(row, col - 1);
    }

    data->updateCell(row, col);

    break;

  case Key_Home:

    if(row > 0)
    {
      data->scroll(-row);
    }
    else // row == 0
    {
      uint oldCol = col;
      col = 0;
      subFocus = 0;
      data->updateCell(0, 0);
      data->updateCell(0, oldCol);
    }

    break;

  case Key_End:

    data->scroll(doc.measureLength(position) / skip - row - 1);

    break;

  case Key_Delete:

    if(subFocus == 0)
    {
      doc[device][pattern][col][row * skip].clear();
    }
    else if(subFocus > 1)
    {
      doc[device][pattern][col][row * skip].setVelocity(127);
    }
    else
    {
      doc[device][pattern][col][row * skip].setRelativity(0);
    }

    doc.setModified();

    data->scroll(spinStep->value());

    break;

  case Key_Plus:

    if(subFocus == 1)
    {
      doc[device][pattern][col][row * skip].setRelativity(1);

      doc.setModified();

      data->scroll(spinStep->value());
    }

    break;

  case Key_Minus:

    if(subFocus == 1)
    {
      doc[device][pattern][col][row * skip].setRelativity(-1);

      doc.setModified();

      data->scroll(spinStep->value());
    }

    break;

  default:

    return;
  }

  event->accept();
}

void Pattern::Data::paintEvent(QPaintEvent* event)
{
  QPainter painter(this);

  painter.setPen(colorGroup().background());
  painter.setBrush(colorGroup().background());

  if(!parent->pattern)
  {
    painter.drawRect(event->rect().intersect(QRect(0, 0, width(), height())));

    return;
  }

  uint patternLength = parent->doc.measureLength(parent->position) / skip;
  const SanetrackerDoc::Pattern& pattern = parent->doc[parent->device][parent->pattern];

  int middle = height() / 2;
  
  const char* hex = "0123456789ABCDEF";
  const char* notes[] = { "C-", "C#", "D-", "D#", "E-", "F-", "F#", "G-", "G#", "A-", "A#", "B-" };
  const char* octaves = "0123456789X";

  painter.setClipRect(3, 3, width() - 6, height() - 6);

  for(int i = -(height() - middle - 11 + 4) / 8; i < (height() - middle + 3) / 8; ++i)
  {
    int x, y;

    x = 4;

    if(i < 0)
      y = middle - 11 + i * 8;
    else if(i == 0)
      y = middle - 4;
    else // i > 0
      y = middle + 3 + i * 8;

    int row = i + parent->row;
    bool validRow = (row >= 0) && (row < static_cast<int>(patternLength));

    TextMode textMode = (row % 4) ? Background : BackgroundDark;

    if(event->rect().intersects(QRect(x, y, 2 * 8, 8)))
    {
      if(!validRow)
      {
        painter.drawRect(x, y, 2 * 8, 8);
      }
      else
      {
        Pattern::drawText(painter, x, y, QString("%1%2").arg(hex[row / 16]).arg(hex[row % 16]), textMode);
      }
    }

    for(uint col = 0; x < width(); ++col)
    {
      const SanetrackerDoc::Pattern::Track& track = pattern[col];

      x = 2 * 7 + 11 + col * (8 * 7 + 5);

      if(event->rect().intersects(QRect(x, y, 7 * 8, 8)))
      {
        if((row % 4) || !validRow)
        {
          painter.setPen(colorGroup().background());
          painter.setBrush(colorGroup().background());
        }
        else // (row % 4) == 0 && validRow
        {
          painter.setPen(colorGroup().background().dark(110));
          painter.setBrush(colorGroup().background().dark(110));
        }

        if(i != 0)
          painter.drawRect(x, y, 7 * 8, 8);
        else
          painter.drawRect(x - 1, y - 2, 7 * 8 + 2, 11);

        if(!validRow)
          continue;

        QString note;
        QString relativity;
        QString velocity0;
        QString velocity1;

        if(track[row * skip].isEmpty())
        {
          note = "---";
          relativity = " ";
          velocity0 = "-";
          velocity1 = "-";
        }
        else if(track[row * skip].isStop())
        {
          note = "[=]";
          relativity = " ";
          velocity0 = "-";
          velocity1 = "-";
        }
        else // !track[row * skip].isEmpty() && !track[row * skip].isStop()
        {
          note = QString("%1%2")
            .arg(notes[track[row * skip].note() % 12])
            .arg(octaves[track[row * skip].note() / 12]);
          relativity = (track[row * skip].relativity() == 1) ? "+"
                     : (track[row * skip].relativity() == 0) ? " "
                                                      : "-";
          velocity0 = QString("%1").arg(hex[track[row * skip].velocity() / 16]);
          velocity1 = QString("%1").arg(hex[track[row * skip].velocity() % 16]);
        }

        bool highlight = parent->active() && (i == 0) && (col == parent->col);

        if(highlight)
        {
          painter.setPen(colorGroup().highlight());
          painter.setBrush(colorGroup().highlight());

          if(parent->subFocus == 0)
            painter.drawRect(x - 1,     middle - 6, 3 * 8 + 1, 11);
          else if(parent->subFocus == 1)
            painter.drawRect(x + 4 * 8, middle - 6, 1 * 8, 11);
          else if(parent->subFocus == 2)
            painter.drawRect(x + 5 * 8, middle - 6, 1 * 8, 11);
          else if(parent->subFocus == 3)
            painter.drawRect(x + 6 * 8, middle - 6, 1 * 8, 11);

          painter.setPen(colorGroup().background());
          painter.setBrush(colorGroup().background());
        }

        Pattern::drawText(painter, x, y, note,
          (highlight && parent->subFocus == 0) ? Highlight : textMode);
        Pattern::drawText(painter, x + 4 * 8, y, relativity,
          (highlight && parent->subFocus == 1) ? Highlight : textMode);
        Pattern::drawText(painter, x + 5 * 8, y, velocity0,
          (highlight && parent->subFocus == 2) ? Highlight : textMode);
        Pattern::drawText(painter, x + 6 * 8, y, velocity1,
          (highlight && parent->subFocus == 3) ? Highlight : textMode);
      }
    }
  }
 
  painter.setClipping(false);

  painter.setClipRect(event->rect());

  painter.setPen(Qt::black);
  painter.drawLine(2, 2, width() - 3, 2);
  painter.drawLine(2, 0, 2, height() - 4);
  painter.drawLine(3, middle - 10, width() - 1, middle - 10);
  painter.drawLine(3, middle - 7, width() - 1, middle - 7);
  painter.drawLine(3, middle + 5, width() - 1, middle + 5);
  painter.drawLine(3, middle + 8, width() - 1, middle + 8);
  painter.drawLine(0, height() - 3, width() - 4, height() - 3);
  painter.drawLine(width() - 3, 0, width() - 3, height() - 3);

  painter.drawLine(2 * 7 + 6,     3, 2 * 7 + 6, height() - 3);
  painter.drawLine(2 * 7 + 6 + 3, 3, 2 * 7 + 6 + 3, height() - 3);

  for(int x = 10 * 7 + 11; x < width(); x += 8 * 7 + 5)
  {
    painter.drawLine(x,     3, x,      height() - 3);
    painter.drawLine(x + 3, 3, x + 3, height() - 3);
  }

  painter.setPen(Qt::gray);
  painter.drawLine(1, 1, width(), 1);
  painter.drawLine(1, 1, 1, height() - 1);
  painter.drawLine(1, middle - 8, width() - 3, middle - 8);
  painter.drawLine(1, middle + 7, width() - 3, middle + 7);
  painter.drawLine(0, height() - 1, width() - 2, height() - 1);
  painter.drawLine(width() - 1, 0, width() - 1, height() - 1);
  
  painter.drawLine(2 * 7 + 6 + 2, 2, 2 * 7 + 6 + 2, height() - 3);

  for(int x = 10 * 7 + 11; x < width(); x += 8 * 7 + 5)
  {
    painter.drawLine(x + 2, 2, x + 2, height() - 3);
  }

  painter.setPen(Qt::white);
  painter.drawLine(0, 0, width() - 2, 0);
  painter.drawLine(0, 0, 0, height() - 2);
  painter.drawLine(0, middle - 9, width() - 3, middle - 9);
  painter.drawLine(0, middle + 6, width() - 3, middle + 6);
  painter.drawLine(1, height() - 2, width() - 2, height() - 2);
  painter.drawLine(width () - 2, 1, width() - 2, height() - 3);
  
  painter.drawLine(2 * 7 + 6 + 1, 1, 2 * 7 + 6 + 1, height() - 3);

  for(int x = 10 * 7 + 11; x < width(); x += 8 * 7 + 5)
  {
    painter.drawLine(x + 1, 1, x + 1, height() - 3);
  }
}

#ifndef WHEEL_DELTA
#define WHEEL_DELTA 120
#endif

void Pattern::Data::wheelEvent(QWheelEvent* event)
{
  scroll(-event->delta() / WHEEL_DELTA);
}

void Pattern::Data::mousePressEvent(QMouseEvent* event)
{
  if(!parent->pattern)
    return;

  int middle = height() / 2;
    
  if(event->y() < middle - 11)
  {
    parent->selectRow = parent->row + (event->y() - (middle - 4)) / 8;
  } else if(event->y() > middle + 3)
  {
    parent->selectRow = parent->row + (event->y() - (middle + 4)) / 8;
  }
  else
    parent->selectRow = parent->row;

  parent->selectCol = (event->x() - (2 * 7 + 11)) / (8 * 7 + 5);

  if(!parent->active())
    emit parent->activated();

  if(event->button() == Qt::RightButton)
    parent->menu->popup(event->globalPos());
}

void Pattern::Data::mouseReleaseEvent(QMouseEvent* /*event*/)
{
}

void Pattern::Data::mouseMoveEvent(QMouseEvent* /*event*/)
{
}

QPixmap* Pattern::fontPixmaps[3][25] = { { 0 } };

const char* Pattern::font[25][9] =
{
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "  xx    ",
    " xxx    ",
    "  xx    ",
    "  xx    ",
    "  xx    ",
    "  xx    ",
    "xxxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "    xx  ",
    "   xx   ",
    "  xx    ",
    " xx     ",
    "xxxxxxx ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "     xx ",
    "   xxx  ",
    "     xx ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
    "xxxxxxx ",
    "     xx ",
    "     xx ",
    "     xx ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "xxxxxxx ",
    "xx      ",
    "xxxxxx  ",
    "     xx ",
    "     xx ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxxx ",
    "xx      ",
    "xxxxxx  ",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "xxxxxxx ",
    "     xx ",
    "     xx ",
    "  xxxx  ",
    "    xx  ",
    "   xx   ",
    "   xx   ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "xx   xx ",
    " xxxxx  ",
    "xx   xx ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "xx   xx ",
    " xxxxxx ",
    "     xx ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "xx   xx ",
    "xxxxxxx ",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "xxxxxx  ",
    "xx   xx ",
    "xx   xx ",
    "xxxxxx  ",
    "xx   xx ",
    "xx   xx ",
    "xxxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "xx      ",
    "xx      ",
    "xx      ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "xxxxx   ",
    "xx  xx  ",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
    "xx   xx ",
    "xxxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "xxxxxxx ",
    "xx      ",
    "xx      ",
    "xxxxx   ",
    "xx      ",
    "xx      ",
    "xxxxxxx ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "xxxxxxx ",
    "xx      ",
    "xx      ",
    "xxxxx   ",
    "xx      ",
    "xx      ",
    "xx      ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    " xxxxx  ",
    "xx   xx ",
    "xx      ",
    "xx xxxx ",
    "xx   xx ",
    "xx   xx ",
    " xxxxx  ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "        ",
    "        ",
    " xxxxx  ",
    "        ",
    "        ",
    "        ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "   xx   ",
    "   xx   ",
    " xxxxxx ",
    "   xx   ",
    "   xx   ",
    "        ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "        ",
    "xxxxxxxx",
    "        ",
    "xxxxxxxx",
    "        ",
    "        ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "  x x   ",
    " xxxxx  ",
    "  x x   ",
    " xxxxx  ",
    "  x x   ",
    "        ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "  x     ",
    "  x     ",
    "  xxx   ",
    "  x x   ",
    "  xxx   ",
    "        ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "        ",
    "   xxxxx",
    "   xx   ",
    "   xxxxx",
    "        ",
    "        ",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "        ",
    "        ",
    "        ",
    "        ",
    "        ",
    "xxxxxxxx",
  },
  {
    "8 7 2 1",
    "x	c #000000",
    "        ",
    "        ",
    "xxxx    ",
    "  xx    ",
    "xxxx    ",
    "        ",
    "        ",
  },
};

