/***************************************************************************
                          sanetrackerdoc.cpp  -  description
                             -------------------
    begin                : Sat Jun  8 15:33:59 CEST 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <iostream>
#include <assert.h>

// include files for Qt
#include <qdir.h>
#include <qregexp.h>
#include <qwidget.h>
#include <qthread.h>
#include <qwidgetlist.h>

// include files for KDE
#include <kapplication.h>
#include <klocale.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <ktempfile.h>
#include <kio/job.h>
#include <kio/netaccess.h>

// application specific includes
#include "sanetrackerdoc.h"
#include "sanetracker.h"
#include "sanetrackerview.h"
#include "mididriver.h"
#include "midifile.h"
#include "midi.h"

const SanetrackerDoc::Pattern::Track::Cell SanetrackerDoc::Pattern::Track::_emptyCell;
const SanetrackerDoc::Pattern::Track       SanetrackerDoc::Pattern::_emptyTrack;
const SanetrackerDoc::Pattern              SanetrackerDoc::Device::_emptyPattern;

SanetrackerDoc::SanetrackerDoc(QObject* parent, const Configuration& config)
  : QObject(parent, 0),
    _config(config)
{
  _devices.setAutoDelete(true);
  _timeBase.position = 0;
  _timeBase.bpm = 120;
  _timeBase.numerator = 4;
  _timeBase.denominator = 2;
  _timeBase.clocks = 24;
  _timeBase.freq = 8;
  _timeBase.next = 0;
}

SanetrackerDoc::~SanetrackerDoc()
{
}

SanetrackerDoc::Pattern::Pattern()
  : _patch(0),
    _tracks(0),
    _trackCount(0)
{
}

SanetrackerDoc::Pattern::~Pattern()
{
  if(_trackCount)
    delete [] _tracks;
}

SanetrackerDoc::Pattern&
SanetrackerDoc::Pattern::operator=(const Pattern& pattern)
{
  if(_trackCount)
    delete [] _tracks;

  _patch = pattern._patch;
  _trackCount = pattern._trackCount;

  _tracks = new Track[_trackCount];

  for(uint i = 0; i < _trackCount; ++i)
    _tracks[i] = pattern._tracks[i];

  return *this;
}

SanetrackerDoc::Pattern::Track::Track()
  : _cellCount(0)
{
}

SanetrackerDoc::Pattern::Track::~Track()
{
  if(_cellCount)
    delete [] _cells;
}

SanetrackerDoc::Pattern::Track&
SanetrackerDoc::Pattern::Track::operator=(const Track& track)
{
  if(_cellCount)
    delete [] _cells;

  _cellCount = track._cellCount;

  _cells = new Cell[_cellCount];

  for(uint i = 0; i < _cellCount; ++i)
    _cells[i] = track._cells[i];

  return *this;
}

SanetrackerDoc::Pattern::Track::Cell&
SanetrackerDoc::Pattern::Track::operator[](uint index)
{
  if(index >= _cellCount)
  {
    uint newCellCount = ((index + 32) / 32) * 32;
    Cell* newCells = new Cell[newCellCount];

    for(uint i = 0; i < _cellCount; ++i)
      newCells[i] = _cells[i];

    if(_cellCount)
      delete [] _cells;

    _cellCount = newCellCount;
    _cells = newCells;
  }

  return _cells[index];
}

const SanetrackerDoc::Pattern::Track::Cell&
SanetrackerDoc::Pattern::Track::operator[](uint index) const
{
  return (index >= _cellCount) ? _emptyCell : _cells[index];
}

SanetrackerDoc::Pattern::Track&
SanetrackerDoc::Pattern::operator[](uint index)
{
  if(index >= _trackCount)
  {
    uint newTrackCount = index + 1;
    Track* newTracks = new Track[newTrackCount];

    for(uint i = 0; i < _trackCount; ++i)
    {
      newTracks[i]._cellCount = _tracks[i]._cellCount;
      newTracks[i]._cells = _tracks[i]._cells;

      _tracks[i]._cellCount = 0;
    }

    if(_trackCount)
      delete [] _tracks;

    _trackCount = newTrackCount;
    _tracks = newTracks;
  }

  return _tracks[index];
}

const SanetrackerDoc::Pattern::Track&
SanetrackerDoc::Pattern::operator[](uint index) const
{
  return (index >= _trackCount) ? _emptyTrack : _tracks[index];
}

void
SanetrackerDoc::Pattern::remove(uint index)
{
  if(index >= _trackCount)
    return;

  if(index == 0 && _trackCount == 1)
  {
    delete [] _tracks;
  }
  else
  {
    Track* newTracks = new Track[_trackCount - 1];

    for(uint i = 0; i < index; ++i)
      newTracks[i] = _tracks[i];

    for(uint i = index; i < _trackCount - 1; ++i)
      newTracks[i] = _tracks[i + 1];

    delete [] _tracks;
    _tracks = newTracks;
  }

  --_trackCount;
}

SanetrackerDoc::EventList::EventList()
{
}

SanetrackerDoc::EventList::EventList(const EventList& events, uint time)
{
  QValueList<Event>::const_iterator i;

  for(i = events._events.begin(); i != events._events.end(); ++i)
  {
    if((*i).time() > time)
      break;

    if((*i).time() == time)
      (*this) += *i;
  }
}

SanetrackerDoc::EventList&
SanetrackerDoc::EventList::operator+=(const Event& event)
{
  QValueList<Event>::iterator i;

  for(i = _events.begin(); i != _events.end(); ++i)
    if((*i).time() > event.time())
      break;

  _events.insert(i, event);

  return *this;
}

uint
SanetrackerDoc::EventList::size() const
{
  return _events.size();
}

SanetrackerDoc::Event&
SanetrackerDoc::EventList::operator[](uint index)
{
  return _events[index];
}

const SanetrackerDoc::Event&
SanetrackerDoc::EventList::operator[](uint index) const
{
  return _events[index];
}

void
SanetrackerDoc::EventList::remove(uint index)
{
  QValueList<Event>::iterator i = _events.begin();

  while(index--)
    ++_events.begin();

  _events.remove(i);
}

SanetrackerDoc::Device::Device(const Configuration& config, uint portIndex,
  uint deviceID, const QString& configName, const QString& configType)
  : _config(config),
    _portIndex(portIndex),
    _deviceID(deviceID),
    _configName(configName),
    _configType(configType),
    _name(configName),
    _type(configType),
    _volume(127)
{
  MIDIDriver::putRT(portIndex, MIDI_CONTROL(deviceID));
  MIDIDriver::putRT(portIndex, MIDI_VOLUME);
  MIDIDriver::putRT(portIndex, _volume);
}

uint
SanetrackerDoc::Device::newPattern() const
{
  uint prev = 0;

  for(QMap<uint, Pattern>::const_iterator i = _patterns.begin();
      i != _patterns.end(); ++i, ++prev)
  {
    if(i.key() != prev + 1)
      return prev + 1;
  }

  return prev + 1;
}

SanetrackerDoc::Pattern&
SanetrackerDoc::Device::operator[](uint index)
{
  return _patterns[index];
}

const SanetrackerDoc::Pattern&
SanetrackerDoc::Device::operator[](uint index) const
{
  QMap<uint, Pattern>::const_iterator i = _patterns.find(index);

  if(i == _patterns.end())
    return _emptyPattern;

  return *i;
}

uint
SanetrackerDoc::Device::size() const
{
  return _patterns.size();
}

void
SanetrackerDoc::Device::remove(uint index)
{
  _patterns.remove(index);
}

uint
SanetrackerDoc::Device::sequence(uint measure) const
{
  Sequence::const_iterator i = _sequence.find(measure);

  if(i == _sequence.end())
    return 0;

  return *i;
}

void
SanetrackerDoc::Device::setSequence(uint measure, uint pattern)
{
  if(pattern)
  {
    _sequence[measure] = pattern;
  }
  else
  {
    Sequence::iterator i = _sequence.find(measure);

    if(i != _sequence.end())
      _sequence.erase(i);
  }
}

void
SanetrackerDoc::Device::clearSequence()
{
  _sequence.clear();
}

void
SanetrackerDoc::Device::setVolume(uint volume)
{
  if(volume > 127)
  {
    kdWarning() << "Volume " << volume << " out of range." << endl;

    return;
  }

  if(volume == _volume)
    return;

  _volume = volume;

  MIDIDriver::putRT(portIndex(), MIDI_CONTROL(deviceID()));
  MIDIDriver::putRT(portIndex(), MIDI_VOLUME);
  MIDIDriver::putRT(portIndex(), volume);
}

bool
SanetrackerDoc::newDocument()
{
  _devices.clear();

  for(uint port = 0; port < _config.portCount(); ++port)
    for(uint channel = 0; channel < MIDI_CHANNELS_MAX; ++channel)
      if(!_config.port(port).device(channel).type().isEmpty())
      {
        _devices.append(
          new Device(
            _config,
            port,
            channel,
            _config.port(port).device(channel).name(),
            _config.port(port).device(channel).type()));
      }

  setModified(false);
  _url.setFileName(i18n("Untitled"));

  // Update all visible widgets
  QWidgetList* widgets = QApplication::allWidgets();

  for(QWidgetListIt i = *widgets; i.current(); ++i)
  {
    if(i.current()->isVisible())
      i.current()->update();
  }

  return true;
}

bool
SanetrackerDoc::openDocument(const KURL& url)
{
  kdDebug() << "Opening file \"" << url.fileName() << "\"." << endl;

  _devices.clear();

  QMap<QString, Device*> devices;

  for(uint port = 0; port < _config.portCount(); ++port)
    for(uint channel = 0; channel < MIDI_CHANNELS_MAX; ++channel)
      if(!_config.port(port).device(channel).type().isEmpty())
      {
        _devices.append(
          new Device(
            _config,
            port,
            channel,
            _config.port(port).device(channel).name(),
            _config.port(port).device(channel).type()));

        devices[_config.port(port).device(channel).name()] = _devices.last();
      }

  setModified(false);

  MIDIFile input;
  QString tmpFile;

  if(url.isLocalFile())
  {
    input.setName(url.directory(false) + url.fileName());
  }
  else
  {
#if KDE_VERSION < KDE_MAKE_VERSION(3,2,0)
    KIO::NetAccess::download(url, tmpFile);
#else
    KIO::NetAccess::download(url, tmpFile, 0);
#endif

    input.setName(tmpFile);
  }

  if(!input.open(IO_ReadOnly))
  {
    // XXX: Give a better error message.
    KMessageBox::error(static_cast<QWidget*>(parent()), i18n("Unable to open file for reading."));

    return false;
  }

  MIDIFile::ChunkVector chunks = input.chunks();

  input.close();

  if(chunks.isEmpty() || chunks[0]->id() != "MThd")
  {
    KMessageBox::error(static_cast<QWidget*>(parent()), i18n("Specified file is not a SMF file (missing MThd chunk)."));

    return false;
  }

  // MIDIFile::MThd* header = static_cast<MIDIFile::MThd*>(chunks[0]);

  QString instrument;
  int     port = -1;

  for(uint i = 1; i < chunks.size(); ++i)
  {
    if(chunks[i]->id() == "MTrk")
    {
      const MIDIFile::MTrk* mtrk = static_cast<MIDIFile::MTrk*>(chunks[i]);
      uint baseTime = (*mtrk->begin()).time();

      uint col = 0;
      
      for(uint time = 0; time < baseTime; time += measureLength(col++));

      unsigned char* cols = 0;
      int      track = -1;
      uint     n = 0;

      bool headChunk = false;

      for(MIDIFile::MTrk::ConstIterator event = mtrk->begin(); event != mtrk->end(); ++event)
      {
        if((*event).data(0) < 0x80) // Meta-event
        {
          switch((*event).data(0))
          {
          case 0x03: // Trackname

            {
              QString temp;

              temp.setLength((*event).size() - 1);

              for(uint i = 0; i < (*event).size() - 1; ++i)
                temp[i] = (*event).data(i + 1);

              track = temp.toInt();

              if(!instrument.isNull())
              {
                devices[instrument]->setSequence(col, track);
              }
            }

            break;

          case 0x04: // Instrument

            {
              instrument.setLength((*event).size() - 1);

              for(uint i = 0; i < (*event).size() - 1; ++i)
                instrument[i] = (*event).data(i + 1);

              if(!devices.contains(instrument))
              {
                // XXX: handle this
                kdDebug() << "Missing instrument \"" << instrument << "\"" << endl;

                break;
              }
            }

            break;

          case 0x09: // Portname

            {
              QString temp;

              temp.setLength((*event).size() - 1);

              for(uint i = 0; i < (*event).size() - 1; ++i)
                temp[i] = (*event).data(i + 1);

              port = temp.toInt();

              headChunk = true;
            }

            break;

          case 0x7F:

            // "Sane" -- Sanetracker event
            if((*event).data(1) == 'S' && (*event).data(2) == 'a'
            && (*event).data(3) == 'n' && (*event).data(4) == 'e')
            {
              // "Cols" -- Column indexes for each note event
              if((*event).data(5) == 'C' && (*event).data(6) == 'o'
              && (*event).data(7) == 'l' && (*event).data(8) == 's')
              {
                cols = new unsigned char[(*event).size() - 9];

                for(uint i = 0; i < (*event).size() - 9; ++i)
                  cols[i] = (*event).data(i + 9);
              }
              else
              // "Patc" -- Virtual patch number
              if((*event).data(5) == 'P' && (*event).data(6) == 'a'
              && (*event).data(7) == 't' && (*event).data(8) == 'c')
              {
                uint patch = ((*event).data(9) << 24)
                           | ((*event).data(10) << 16)
                           | ((*event).data(11) << 8)
                           | ((*event).data(12));

                (*devices[instrument])[track].setPatch(patch);
              }
            }

            break;
          }
        }
        else // Not meta-event
        {
          switch((*event).data(0) & 0xF0)
          {
          case MIDI_NOTE_ON(0):

            if(!cols)
            {
              Event e((*event).time());

              e.put(MIDI_NOTE_ON(0));
              e.put((*event).data(1));
              e.put((*event).data(2));

              (*this)[(*event).data(0) & 0x0F].events() += e;
            }
            else
            {
              if((*event).data(2) == 0)
                break;

              if(((*event).time() < baseTime) || (*event).time() >= baseTime + measureLength(col))
              {
                kdError() << "\"Note on\" event outside of pattern " << (*event).time() << " " << baseTime << endl;

                break;
              }

              Pattern::Track::Cell& cell = (*devices[instrument])[track][cols[n]][(*event).time() - baseTime];

              cell.setNote((*event).data(1));
              cell.setVelocity((*event).data(2));
              cell.setRelativity(0);
              // XXX: if(cols[n + 1]) handle relativity

              n += 2;
            }

            break;

          case MIDI_NOTE_OFF(0):

            if(!cols)
            {
              Event e((*event).time());

              e.put(MIDI_NOTE_OFF(0));
              e.put((*event).data(1));
              e.put((*event).data(2));

              (*this)[(*event).data(0) & 0x0F].events() += e;
            }
            else
            {
              if(((*event).time() < baseTime) || (*event).time() >= baseTime + measureLength(col))
              {
                kdError() << "\"Note off\" event outside of pattern " << (*event).time() << " " << baseTime << endl;

                break;
              }

              Pattern::Track::Cell& cell = (*devices[instrument])[track][cols[n]][(*event).time() - baseTime];

              cell.setStop();

              n++;
            }

            break;

          case MIDI_CONTROL(0):

            switch((*event).data(1))
            {
            case MIDI_VOLUME:

              if(headChunk)
                devices[instrument]->setVolume((*event).data(2));

              break;
            }

            break;
          }
        }
      }

      if(cols)
        delete [] cols;
    }
  }

  if(!tmpFile.isEmpty())
    KIO::NetAccess::removeTempFile(tmpFile);

  // Update all visible widgets
  QWidgetList* widgets = QApplication::allWidgets();

  for(QWidgetListIt i = *widgets; i.current(); ++i)
  {
    if(i.current()->isVisible())
      i.current()->update();
  }

  delete widgets; 

  _url = url;

  return true;
}

bool
SanetrackerDoc::saveDocument(const KURL& url)
{
  MIDIFile output;

  if(url.isLocalFile())
  {
    output.setName(url.directory(false) + url.fileName());

    if(!output.open(IO_WriteOnly))
    {
      // XXX: Give a better error message.
      _error = i18n("Unable to open file for writing");

      return false;
    }
  }
  else
  {
    // XXX: Fix this!
    _error = i18n("Sanetracker doesn't support this filesystem yet.");

    return false;
  }

  uint patternCount = 0;
  for(uint i = 0; i < size(); ++i)
    patternCount += (*this)[i].size();

  output.writeChunk(MIDIFile::MThd(MIDIFile::MThd::Patterns, patternCount, 24));

  MIDIFile::MTrk chunk;

  for(uint i = 0; i < size(); ++i)
  {
    QMap<uint, uint> activeNotes;
    QMap<uint, bool> trackWritten;
    uint currentPatch = 0;

    Device::Sequence::const_iterator seq;

    MIDIFile::MTrk chunk;
    chunk.put(MIDIFile::MTrk::PortName(QString::number((*this)[i].portIndex()),
              0));
    chunk.put(MIDIFile::MTrk::Instrument((*this)[i].configName(), 0));

    chunk.put(MIDIFile::MTrk::EndOfTrack(0));

    MIDIFile::MTrk::Event volume(0);

    volume.put(MIDI_CONTROL((*this)[i].deviceID()));
    volume.put(MIDI_VOLUME);
    volume.put((*this)[i].volume());
    chunk.put(volume);

    output.writeChunk(chunk);

    for(seq = (*this)[i]._sequence.begin(); seq != (*this)[i]._sequence.end(); ++seq)
    {
      const Pattern& pattern = (*this)[i][seq.data()];

      uint baseTime = 0;

      for(uint j = 0; j < seq.key(); ++j)
        baseTime += measureLength(j);

      MIDIFile::MTrk chunk;

      chunk.put(MIDIFile::MTrk::TrackName(QString::number(seq.data()), baseTime));

      uint patternLength = measureLength(seq.key());

      if(trackWritten.find(seq.data()) == trackWritten.end())
      {
        trackWritten[seq.data()] = true;

        QValueVector<unsigned char> info;

        // "SaneCols"
        info.push_back('S'); info.push_back('a'); info.push_back('n'); info.push_back('e');
        info.push_back('C'); info.push_back('o'); info.push_back('l'); info.push_back('s');

        // Write which events belongs to which column, and whether parameters
        // are relative or absolute.
        for(uint row = 0; row < patternLength; ++row)
        {
          for(uint col = 0; col < pattern.size(); ++col)
          {
            if(pattern[col][row].isStop())
            {
              info.push_back(col);
            }
            else if(!pattern[col][row].isEmpty())
            {
              info.push_back(col);
              info.push_back(pattern[col][row].relativity() ? 1 : 0);
            }
          }
        }

        chunk.put(MIDIFile::MTrk::Proprietary(info.begin(), info.size(), baseTime));

        info.clear();

        // "SanePatc"
        info.push_back('S'); info.push_back('a'); info.push_back('n'); info.push_back('e');
        info.push_back('P'); info.push_back('a'); info.push_back('t'); info.push_back('c');

        // Write virtual patch number
        info.push_back(pattern.patch() >> 24);
        info.push_back((pattern.patch() >> 16) & 0xFF);
        info.push_back((pattern.patch() >> 8) & 0xFF);
        info.push_back(pattern.patch() & 0xFF);

        chunk.put(MIDIFile::MTrk::Proprietary(info.begin(), info.size(), baseTime));
      }

      if(pattern.patch() != 0 && pattern.patch() != currentPatch)
      {
        uint physPatch = _config.instrument((*this)[i].configType()).patchList()[pattern.patch() - 1];

        MIDIFile::MTrk::Event event(baseTime);

        event.put(MIDI_CONTROL((*this)[i].deviceID()));
        event.put(MIDI_BANK_SELECT_FINE);
        event.put((physPatch >> 7) & 0x7F);
        chunk.put(event);

        event.clear();
        event.put(MIDI_CONTROL((*this)[i].deviceID()));
        event.put(MIDI_BANK_SELECT);
        event.put(physPatch >> 14);
        chunk.put(event);

        event.clear();
        event.put(MIDI_PATCH_CHANGE((*this)[i].deviceID()));
        event.put(physPatch & 0x7F);
        chunk.put(event);

        currentPatch = pattern.patch();
      }

      for(uint row = 0; row < patternLength; ++row)
      {
        for(uint col = 0; col < pattern.size(); ++col)
        {
          if(pattern[col][row].isEmpty())
            continue;

          MIDIFile::MTrk::Event event(baseTime + row);

          if(pattern[col][row].isStop())
          {
            event.put(MIDI_NOTE_OFF((*this)[i].deviceID()));
            event.put(activeNotes[col]);
            event.put(127);

            chunk.put(event);

            activeNotes.erase(col);

            continue;
          }

          if(activeNotes.find(col) != activeNotes.end())
          {
            event.put(MIDI_NOTE_ON((*this)[i].deviceID()));
            event.put(activeNotes[col]);
            event.put(0);

            chunk.put(event);

            activeNotes.erase(col);
          }

          event.clear();
          event.put(MIDI_NOTE_ON((*this)[i].deviceID()));
          event.put(pattern[col][row].note());
          // XXX: handle relativity
          event.put(pattern[col][row].velocity());

          chunk.put(event);

          activeNotes[col] = pattern[col][row].note();
        }
      }

      chunk.put(MIDIFile::MTrk::EndOfTrack(baseTime + measureLength(seq.key())));
      output.writeChunk(chunk);
    }
  }

  output.close();

  _url = url;

  setModified(false);

  return true;
}

QString
SanetrackerDoc::error() const
{
  return _error;
}

uint
SanetrackerDoc::songLength() const
{
  uint length = 0;

  for(uint i = 0; i < size(); ++i)
  {
    Device::Sequence::ConstIterator p;

    for(p = (*this)[i]._sequence.begin(); p != (*this)[i]._sequence.end(); ++p)
    {
      if(p.key() + 1 > length)
        length = p.key() + 1;
    }
  }

  return length;
}

SanetrackerDoc::Device&
SanetrackerDoc::operator[](uint index)
{
  return *_devices.at(index);
}

const SanetrackerDoc::Device&
SanetrackerDoc::operator[](uint index) const
{
  return *const_cast<const SanetrackerDoc::Device*>(_devices.at(index));
}

uint
SanetrackerDoc::size() const
{
  return _devices.count();
}

void
SanetrackerDoc::insertMeasure(uint index)
{
  for(uint i = 0; i < size(); ++i)
  {
    Device::Sequence& seq = (*this)[i]._sequence;

    int max = -1;

    QMap<uint, uint>::ConstIterator p;

    for(p = seq.begin(); p != seq.end(); ++p)
    {
      if(static_cast<int>(p.key()) > max)
        max = p.key();
    }

    for(; max >= static_cast<int>(index); --max)
    {
      if(seq.find(max) == seq.end())
        continue;

      seq[max + 1] = seq[max];
      seq.erase(max);
    }
  }

  setModified();
}

void
SanetrackerDoc::removeMeasure(uint index)
{
  for(uint i = 0; i < size(); ++i)
  {
    Device::Sequence& seq = (*this)[i]._sequence;

    int max = -1;

    QMap<uint, uint>::ConstIterator p;

    for(p = seq.begin(); p != seq.end(); ++p)
    {
      if(static_cast<int>(p.key()) > max)
        max = p.key();
    }

    seq.erase(index);

    for(uint col = index; static_cast<int>(col) < max; ++col)
    {
      if(seq.find(col + 1) == seq.end())
        continue;

      seq[col] = seq[col + 1];
      seq.erase(col + 1);
    }
  }

  setModified();
}

uint
SanetrackerDoc::measureLength(uint index) const
{
  const TimeBase* time = &_timeBase;

  while(time->next)
  {
    if(time->next->position > index)
      break;

    time = time->next;
  }

  // XXX: get time->freq in here
  return (time->numerator * 4) / (1 << time->denominator) * time->clocks;
}

void
SanetrackerDoc::setModified(bool modified)
{
  if(_modified == modified)
    return;

  _modified = modified;

  emit this->modified(modified);
}

void
SanetrackerDoc::reconfigure()
{
  if(!isModified())
  {
    newDocument();

    return;
  }

  // XXX: Test whether any devices have been removed
}

