/***************************************************************************
                          sequence.cpp  -  description
                             -------------------
    begin                : Sun Jun 9 2002
    copyright            : (C) 2002 by Morten Hustveit
    email                : mortehu@ifi.uio.no
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

// include files for Qt
#include <qscrollbar.h>
#include <qpainter.h>
#include <qlayout.h>
#include <qfontmetrics.h>

// include files for KDE
#include <kapplication.h>
#include <kglobalsettings.h>
#include <kpopupmenu.h>
#include <klocale.h>

#include "sequence.h"
#include "sanetrackerdoc.h"
#include "sanetrackerview.h"
#include "scheduler.h"

Sequence::Sequence(QWidget* parent, SanetrackerDoc* doc, const Scheduler& scheduler)
  : QWidget(parent),
    scheduler(scheduler)
{
  this->doc = doc;

  // Determine ideal width and height for labels (depends on resolution and
  // installed fonts)

  QFontMetrics test(KGlobalSettings::generalFont());
  labelWidth = test.width("888") + 3;
  labelHeight = test.height() + 2;

  labels = new Labels(this);
  posLabels = new PosLabels(this);
  data = new Data(this);
  scrollVert = new QScrollBar(Qt::Vertical, this);
  scrollHoriz = new QScrollBar(Qt::Horizontal, this);
  slotHorizReleased();

  layout = new QGridLayout(this, 3, 3);
  layout->addWidget(labels, 1, 0);
  layout->addWidget(posLabels, 0, 1);
  layout->addWidget(data, 1, 1);
  layout->addMultiCellWidget(scrollVert, 0, 1, 2, 2);
  layout->addMultiCellWidget(scrollHoriz, 2, 2, 0, 1);

  menu = new KPopupMenu(data, "Sequence menu");
  menu->insertItem(i18n("&Insert column before cursor"), this, SLOT(insertColumn()), Key_Insert);
  menu->insertItem(i18n("&Remove column"), this, SLOT(removeColumn()));
  menu->insertItem(i18n("&Clear row"), this, SLOT(clearRow()));

  connect(scrollHoriz, SIGNAL(valueChanged(int)), this, SLOT(slotHorizScroll(int)));
  connect(scrollHoriz, SIGNAL(sliderReleased()), this, SLOT(slotHorizReleased()));
  connect(scrollVert, SIGNAL(valueChanged(int)), this, SLOT(slotVertScroll(int)));

  currentRow = 0;
  currentCol = 0;

  setWFlags(Qt::WResizeNoErase);
  setWFlags(Qt::WRepaintNoErase);
}

Sequence::Labels::Labels(Sequence* parent) : QWidget(parent), parent(parent)
{
  setFocusPolicy(QWidget::StrongFocus);
  setFixedWidth(150);
}

Sequence::PosLabels::PosLabels(Sequence* parent) : QWidget(parent), parent(parent)
{
  setFixedHeight(parent->labelHeight);
}

Sequence::Data::Data(Sequence* parent) : QWidget(parent), parent(parent)
{
  setFocusPolicy(QWidget::StrongFocus);

  setWFlags(Qt::WResizeNoErase);
  setWFlags(Qt::WRepaintNoErase);
}

Sequence::~Sequence()
{
  delete scrollVert;
  delete scrollHoriz;
}

void Sequence::positionChanged()
{
  posLabels->update();
}

void Sequence::reconfigure()
{
}

void Sequence::setActive(bool active)
{
  if(active == _active)
    return;

  _active = active;
  
  QPainter painter(data);

  data->paintCell(painter, currentRow, currentCol);
}

void Sequence::resizeEvent(QResizeEvent* event)
{
  uint rowSpace = data->height() / labelHeight - 1;

  if(rowSpace < doc->size())
    scrollVert->setMaxValue(doc->size() - rowSpace);
  else
    scrollVert->setMaxValue(0);

  QWidget::resizeEvent(event);
}

void Sequence::Labels::paintRow(QPainter& painter, uint row)
{
  uint y = row - parent->scrollVert->value();

  QColor color = (row == parent->currentRow) ? QColor(0xCFCFFF) : QColor(0xDFEFFF);

  if(row & 1)
    color = color.dark(105);

  painter.setFont(KGlobalSettings::generalFont());
  painter.setPen(Qt::black);
  painter.setBrush(color);

  painter.drawRect(-1, y * parent->labelHeight, width() + 1, parent->labelHeight + 1);
  painter.drawText(1, 1 + y * parent->labelHeight, width() - 2, parent->labelHeight, 0, (*parent->doc)[row].name());
}

void Sequence::Labels::paintEvent(QPaintEvent* event)
{
  QPainter painter(this);

  uint row = parent->scrollVert->value();

  for(uint y = 0; y <= height() / parent->labelHeight && row < parent->doc->size(); ++y, ++row)
  {
    if(event->rect().intersects(
        QRect(0, y * parent->labelHeight,
          width(), parent->labelHeight + 1)))
      paintRow(painter, row);
  }
}

void Sequence::Labels::mousePressEvent(QMouseEvent* event)
{
  emit parent->activated();

  if(event->button() == Qt::LeftButton)
  {
    uint prevRow = parent->currentRow;

    parent->currentRow = parent->scrollVert->value() + event->y() / parent->labelHeight;

    if(parent->currentRow >= parent->doc->size())
    {
      parent->currentRow = prevRow;

      return;
    }

    if(prevRow != parent->currentRow)
    {
      QPainter painter(this);

      paintRow(painter, prevRow);
      paintRow(painter, parent->currentRow);

      QPainter painter2(parent->data);

      parent->data->paintCell(painter2, prevRow, parent->currentCol);
      parent->data->paintCell(painter2, parent->currentRow, parent->currentCol);
    }
  }
}

void Sequence::Data::paintCell(QPainter& painter, uint row, uint col)
{
  if(row >= parent->doc->size())
    return;

  uint x = col - parent->scrollHoriz->value();
  uint y = row - parent->scrollVert->value();

  QColor color;
  if(row == parent->currentRow && col == parent->currentCol)
  {
    if(parent->active())
    {
      color = QColor(0xDFDFFF);
    }
    else
    {
      color = QColor(0xEFE8DF);
    }
  }
  else
  {
    color = QColor(0xFFEFBF);

    if(!(row & 1))
      color = color.light(110);
    if(col & 1)
      color = color.light(105);
  }

  painter.setFont(KGlobalSettings::generalFont());
  painter.setPen(Qt::black);
  painter.setBrush(color);

  painter.drawRect(
    x * parent->labelWidth - 1, y * parent->labelHeight,
    parent->labelWidth + 1, parent->labelHeight + 1);

  int pattern;

  if(0 != (pattern = (*parent->doc)[row].sequence(col)))
    painter.drawText(
      x * parent->labelWidth + 2, 1 + y * parent->labelHeight, // position
      parent->labelWidth - 4, parent->labelHeight,         // size
      Qt::AlignRight,                                      // flags
      QString("%1").arg(pattern));
}

void Sequence::Data::mousePressEvent(QMouseEvent* event)
{
  emit parent->activated();

  uint prevRow = parent->currentRow;
  uint prevCol = parent->currentCol;

  parent->currentRow = parent->scrollVert->value() + event->y() / parent->labelHeight;

  if(parent->currentRow >= parent->doc->size())
  {
    parent->currentRow = prevRow;

    return;
  }

  parent->currentCol = parent->scrollHoriz->value() + event->x() / parent->labelWidth;

  if(prevRow != parent->currentRow || prevCol != parent->currentCol)
  {
    if(prevRow != parent->currentRow)
    {
      QPainter painter(parent->labels);

      parent->labels->paintRow(painter, prevRow);
      parent->labels->paintRow(painter, parent->currentRow);
    }

    QPainter painter(this);

    paintCell(painter, prevRow, prevCol);
    paintCell(painter, parent->currentRow, parent->currentCol);
  }

  emit parent->selectionChanged(parent->currentRow, (*parent->doc)[parent->currentRow].sequence(parent->currentCol), parent->currentCol);

  if(event->button() == Qt::RightButton)
    parent->menu->popup(event->globalPos());
}

void Sequence::Data::paintEvent(QPaintEvent* event)
{
  QPainter painter(this);

  uint rowOffset = parent->scrollVert->value();
  uint colOffset = parent->scrollHoriz->value();

  uint y;

  for(y = 0; y <= height() / parent->labelHeight && y + rowOffset < parent->doc->size(); ++y)
    for(uint x = 0; x <= width() / parent->labelWidth; ++x)
    {
      if(event->rect().intersects(
          QRect(x * parent->labelWidth, y * parent->labelHeight,
            parent->labelWidth + 1, parent->labelHeight + 1)))
        paintCell(painter, y + rowOffset, x + colOffset);
    }

  if(y * parent->labelHeight + 1 < static_cast<uint>(height()))
    erase(0, y * parent->labelHeight + 1, width(), height() - y * parent->labelHeight - 1);
}

void Sequence::PosLabels::paintEvent(QPaintEvent* /*event*/)
{
  QPainter painter(this);

  painter.setFont(KGlobalSettings::generalFont());

  uint m = parent->scrollHoriz->value();

  for(uint j = 0;
      j < static_cast<uint>(width()) / parent->labelWidth + 1; ++j, ++m)
  {
    if(parent->scheduler.isPlaying() && m == parent->scheduler.measure())
      painter.setPen(Qt::red);

    painter.drawText(
      j * parent->labelWidth, 1,               // position
      parent->labelWidth, parent->labelHeight, // size
      Qt::AlignLeft,                           // flags
      QString("%1").arg(m + 1));

     if(parent->scheduler.isPlaying() && m == parent->scheduler.measure())
       painter.setPen(Qt::black);
   }
}

void Sequence::insertColumn()
{
  doc->insertMeasure(currentCol);

  doc->setModified();

  data->update();
}

void Sequence::removeColumn()
{
  doc->removeMeasure(currentCol);

  doc->setModified();

  data->update();
}

void Sequence::clearRow()
{
  (*doc)[currentRow].clearSequence();

  doc->setModified();

  data->update();
}

void Sequence::keyPressed(QKeyEvent* event)
{
  if(!doc->size())
    return;

  uint pattern = (*doc)[currentRow].sequence(currentCol);

  uint prevRow = currentRow;
  uint prevCol = currentCol;

  switch(event->key())
  {
  case Key_Home:

    currentCol = 0;

    break;

  case Key_End:

    int songLength;

    songLength = doc->songLength();

    if(songLength == 0)
      songLength = 1;

    currentCol = songLength - 1;

    break;

  case Key_PageUp:

    if(currentRow > data->height() / labelHeight)
      currentRow -= data->height() / labelHeight;
    else
      currentRow = 0;

    break;

  case Key_PageDown:

    if(currentRow < doc->size() - data->height() / labelHeight - 1)
      currentRow += data->height() / labelHeight;
    else
      currentRow = doc->size() - 1;

    break;

  case Key_Up:

    if(currentRow > 0)
      currentRow--;

    break;

  case Key_Down:

    if(currentRow < doc->size() - 1)
      currentRow++;

    break;

  case Key_Left:

    if(currentCol > 0)
      currentCol--;

    break;

  case Key_Right:

    currentCol++;

    break;

  case Key_Delete:

    (*doc)[currentRow].setSequence(currentCol, 0);

    doc->setModified();

    break;

  case Key_Backspace:

    (*doc)[currentRow].setSequence(currentCol, pattern / 10);

    doc->setModified();

    break;

#define HANDLE_DIGIT(n) \
  case ('0' + n): \
    if(pattern < 100) \
    { \
      (*doc)[currentRow].setSequence(currentCol, pattern * 10 + n); \
      doc->setModified(); \
    } \
    break

  HANDLE_DIGIT(0); HANDLE_DIGIT(1); HANDLE_DIGIT(2); HANDLE_DIGIT(3);
  HANDLE_DIGIT(4); HANDLE_DIGIT(5); HANDLE_DIGIT(6); HANDLE_DIGIT(7);
  HANDLE_DIGIT(8); HANDLE_DIGIT(9);

  default:

    return;
  }

  if(currentRow < static_cast<uint>(scrollVert->value()))
  {
    scrollVert->setValue(currentRow);
  }
  else if(currentRow >= scrollVert->value() + data->height() / labelHeight)
  {
    scrollVert->setValue(currentRow - data->height() / labelHeight + 1);
  }

  if(currentCol < static_cast<uint>(scrollHoriz->value()))
  {
    scrollHoriz->setValue(currentCol);
  }
  else if(currentCol >= scrollHoriz->value() + data->width() / labelWidth)
  {
    if(static_cast<uint>(scrollHoriz->maxValue())
       < currentCol - static_cast<uint>(data->width()) / labelWidth + 1)
      scrollHoriz->setMaxValue(currentCol + 16);

    scrollHoriz->setValue(currentCol - data->width() / labelWidth + 1);
  }

  if(prevRow != currentRow || prevCol != currentCol)
  {
    QPainter painter(data);

    data->paintCell(painter, prevRow, prevCol);
    data->paintCell(painter, currentRow, currentCol);
    
    emit selectionChanged(currentRow, (*doc)[currentRow].sequence(currentCol), currentCol);
  }
  else if(pattern != (*doc)[currentRow].sequence(currentCol))
  {
    QPainter painter(data);

    data->paintCell(painter, currentRow, currentCol);
    
    emit selectionChanged(currentRow, (*doc)[currentRow].sequence(currentCol), currentCol);
  }

  if(prevRow != currentRow)
  {
    QPainter painter(labels);

    labels->paintRow(painter, prevRow);
    labels->paintRow(painter, currentRow);
  }

  event->accept();
}

void Sequence::slotHorizScroll(int val)
{
  if(!scrollHoriz->draggingSlider())
    slotHorizReleased();

  static int prevVal;

  data->scroll((prevVal - val) * labelWidth, 0);
  posLabels->scroll((prevVal - val) * labelWidth, 0);

  prevVal = val;
}

void Sequence::slotHorizReleased()
{
  int val = scrollHoriz->value();
  int songLength = doc->songLength();

  if(val > scrollHoriz->maxValue() - 16)
  {
    scrollHoriz->setMaxValue(val + 16);
  }
  else if(scrollHoriz->maxValue() > songLength + 16
  && val < scrollHoriz->maxValue() - 16)
  {
    scrollHoriz->setMaxValue(((songLength > val) ? songLength : val) + 16);
  }
}

void Sequence::slotVertScroll(int val)
{
  static int prevVal;

  data->scroll(0, (prevVal - val) * labelHeight);
  labels->scroll(0, (prevVal - val) * labelHeight);

  prevVal = val;
}

#ifndef WHEEL_DELTA
#define WHEEL_DELTA 120
#endif

void Sequence::wheelEvent(QWheelEvent* event)
{
  scrollVert->setValue(scrollVert->value() - event->delta() / WHEEL_DELTA);
}

void Sequence::slotNextPattern()
{
  ++currentCol;

  if(currentCol >= scrollHoriz->value() + data->width() / labelWidth)
  {
    if(static_cast<uint>(scrollHoriz->maxValue())
       < currentCol - static_cast<uint>(data->width()) / labelWidth + 1)
      scrollHoriz->setMaxValue(currentCol + 16);

    scrollHoriz->setValue(currentCol - data->width() / labelWidth + 1);
  }

  if(!(*doc)[currentRow].sequence(currentCol))
    (*doc)[currentRow].setSequence(currentCol, (*doc)[currentRow].newPattern());

  QPainter painter(data);

  data->paintCell(painter, currentRow, currentCol - 1);
  data->paintCell(painter, currentRow, currentCol);
    
  emit selectionChanged(currentRow, (*doc)[currentRow].sequence(currentCol), currentCol);
}

void Sequence::slotPrevPattern()
{
  if(!currentCol)
    return;

  --currentCol;

  if(currentCol >= scrollHoriz->value() + data->width() / labelWidth)
  {
    if(static_cast<uint>(scrollHoriz->maxValue())
       < currentCol - static_cast<uint>(data->width()) / labelWidth + 1)
      scrollHoriz->setMaxValue(currentCol + 16);

    scrollHoriz->setValue(currentCol - data->width() / labelWidth + 1);
  }

  if(!(*doc)[currentRow].sequence(currentCol))
    (*doc)[currentRow].setSequence(currentCol, (*doc)[currentRow].newPattern());

  QPainter painter(data);

  data->paintCell(painter, currentRow, currentCol + 1);
  data->paintCell(painter, currentRow, currentCol);
    
  emit selectionChanged(currentRow, (*doc)[currentRow].sequence(currentCol), currentCol);
}


